* - CALC.PRG -
*
* This program displays a calculator and allows the user
* to calculate a number and then return the value as
* input to the current field. This program is designed
* to emulate a "10-key adding machine" type calculator
* used by many accountants.
*
* >> This program is a highly modified version of
* >> "calculat.prg" supplied with the DBman distribution.
*
* Notes:
*  o The calculator does not use operator
*    precedence.
*  o Mistakes in the number being input can
*    be corrected by using BackSpace.
*  o The percentage operator is used as a
*    mark up/down function.
*  o The number of decimal places and the
*    overall length of the result can be
*    fixed to accomodate the requirements
*    of the current field.
*  o The calculator can be placed anywhere on
*    the screen to accomodate viewing. The
*    "row/col" coordinates are adjusted to
*    prevent the calculator window from being
*    positioned out of screen bounds.
*  o The calculator utilizes a memory for
*    storing a constant or the result of
*    a calculation.
*
* Examples:
*
*    keystrokes       result    comment
*    -----------      ------    -------------------------
*    5 + 25 %           6.25    mark-up
*    5 - 25 %           3.75    mark-down
*    5 * 25 %           6.25    mark-up
*    6.25 / 25 %        5.00    normalizing-reduction
*    5 + + + =         20.00    repeating number input
*    25 - 5 - - =      10.00    "
*    5 * * * =        625.00    "
*    5 + + - 3 =       12.00    "
*    5 * 5 s =        -25.00    change sign
*+-------------------------------------------------------------+
* Syntax:
*     DO CALC WITH nn, row, col, size, dsize
* Where:
*     "nn". . . . .is the number of decimal places to display.
*     "row, col". .is upper left corner of the calculator.
*     "size". . . .is 0 or 1 for small or large calculator.
*     "dsize" . . .is the size of the display.
*+-------------------------------------------------------------+
* 03-17-93 jhd :-)
*
parameter dx, rw, cl, sz, dz
dfl = "calc"
erase &dfl..pro
*
* save current context
*
save_esc = iifc(set('escape'), 'on', 'off')
save_db3 = iifc(set('db3'), 'on', 'off')
save_vid = vsay()
wna = pushwind()
*
* initialize dbman
*
set procedure to &dfl
set escape off
set db3 off
*
* set defaults
*
if type('dx') <> "N"
  dx = 2
endif
if type('rw') <> "N"
  rw = 3
endif
if type('cl') <> "N"
  cl = 5
endif
if type('sz') <> "N"
  sz = 0
endif
if type('dz') <> "N"
  dz = 11
endif
memory = 0
error = .n.
vrev = 112
vnml = 7
chr = 'c'
*
* adjust calc size option
*
rw = max(rw, 1)
cl = max(cl, 1)
hrw = min(rw, 6)
hcl = min(cl, 57)
dx = max(0, min((dz - 3), dx))
if sz = 0
  rw = min(rw, 10)
  cl = min(cl, 61)
  dz = max((dx + 3), min(11, dz))
  do smcalc
else
  rw = min(rw, 5)
  cl = min(cl, 55)
  dz = max((dx + 3), min(15, dz))
  do bgcalc
endif
rpos = dz + 4
emsg = dupchar(asc('*'), dz)
rmax = val(dupchar(asc('9'), (dz - 2))) / 10^dx
*
* backspace
*
do while .t.
  do case
    case chr $ 'bB' and not error
      if numode
        number = left(number, (len(number) - 1))
        if len(number) = 0
          if pending
            number = str(accu, dz, dx)
          else
            number = result
          endif
          numode = .n.
          @ 0, 3 say number
        else
          @ 0, 3 say rtjust(number, dz)
        endif
      endif
*
* concatenate number
*
    case chr $ '0123456789.' and not error
      if numode
        if val(number + chr) < rmax
          if len(number) < (dz - 1)
            if chr = '.'
              if not '.' $ number
                number = number + chr
              endif
            else
              number = number + chr
            endif
          endif
        endif
      else
        numode = .y.
        number = chr
      endif
      @ 0, 3 say rtjust(number, dz)
*
* calc operation result
*
    case chr = '=' and not error
      if pending
        @ 0, rpos say chr
        do opercalc
        if not error
          @ 0, 3 say result
          pending = .n.
          numode = .n.
        endif
      endif
*
* calc percent result
*
    case chr = '%' and not error
      if pending
        @ 0, rpos say chr
        do pcntcalc
        if not error
          @ 0, 3 say result
          pending = .n.
          numode = .n.
        endif
      endif
*
* calc according to operator
*
    case chr $ '+-*/' and not error
      @ 0, rpos say chr
      if pending
        do opercalc
        if not error
          @ 0, 3 say result
          operation = chr
          numode = .n.
        endif
      else
        if numode
          accu = val(number)
       	  numode = .n.
        else
          accu = val(result)
        endif
        operation = chr
        pending = .y.
      endif
*
* change sign
*
    case chr $ 'sS' and not error
      if numode
        numode = .n.
        temp = -val(number)
        number = str(temp, dz, dx)
        result = number
        @ 0, 3 say result
        @ 0, rpos say chr
      else
        if not pending
          accu = -accu
          result = str(accu, dz, dx)
          @ 0, 3 say result
          @ 0, rpos say chr
        endif
      endif
*
* store to memory
*
    case chr $ 'mM' and not error
      if numode
        numode = .n.
        result = number
        memory = val(result)
      else
        if not pending
          memory = val(result)
        endif
      endif
      if numode or not pending
        @ 0, rpos say chr
        if memory = 0
          @ 0, 1 say ' '
        else
          @ 0, 1 say 'M'
        endif
      endif
*
* recall from memory
*
    case chr $ 'rR' and not error
      @ 0, rpos say chr
      if pending
        number = str(memory, dz, dx)
        @ 0, 3 say number
      else
        result = str(memory, dz, dx)
        @ 0, 3 say result
      endif
      numode = .n.
*
* clear
*
    case chr $ 'cC'
      pending = .n.
      numode = .n.
      error = .n.
      accu = 0
      result = str(accu, dz, dx)
      @ 0, 3 say result
      @ 0, rpos say chr
      number = result
*
* defer final result to field input
*
    case chr $ 'eE' and not error
      set input to result
      exit
*
* quit
*
    case chr $ 'qQ'
      exit
*
* user help
*
    case chr $ 'hH'
      do dumbuser
*
* input or calc error
*
    otherwise
      beep
  endcase
*
* keyboard input
*
  set console off
  wait to chr
  set console on
  do case
    case lastkey() = 8
      chr = 'b'
    case lastkey() = 13
      chr = '='
    case lastkey() = 27
      chr = 'q'
    case lastkey() = 315
      chr = 'h'
  endcase
enddo
*
* restore settings and close current procedure table
*
popwind(wna)
erase | *** remove if necessary ***
set say video to save_vid
set escape &save_esc
set db3 &save_db3
close procedure
return
*
* calculate result for percentage
*
procedure pcntcalc
  temp = abs(val(number) / 100)
  do case
    case operation = '+'
      accu = accu + accu * temp
    case operation = '-'
      accu = accu - accu * temp
    case operation = '*'
      accu = accu * (1 + temp)
    case operation = '/'
      accu = accu / (1 + temp)
  endcase
  do rangeck
return
*
* calculate result for operation
*
procedure opercalc
  temp = val(number)
  do case
    case operation = '+'
      accu = accu + temp
    case operation = '-'
      accu = accu - temp
    case operation = '*'
      accu = accu * temp
    case operation = '/'
      if temp = 0
        @ 0, 3 say emsg
        error = .y.
        return
      endif
      accu = accu / temp
  endcase
  do rangeck
return
*
* check for overflow
*
procedure rangeck
  accu = round(accu, dx)
  if abs(accu) > rmax
    @ 0, 3 say emsg
    error = .y.
  else
    result = str(accu, dz, dx)
  endif
return
*
* display help
*
procedure dumbuser
  wnb = pushwind()
  frame(1)
  set say video to vnml
  window(hrw, hcl, (hrw + 16), (hcl + 20))
  erase
  @  0, 1 say  'Available keys:'
  @  2, 0 say  center('digits   .0-9', 20)
  @  3, 0 say  center('+ add   - sub', 20)
  @  4, 0 say  center('* mul   / div', 20)
  @  6, 3 say  '% percent'
  @  7, 3 say  '= or <enter>'
  @  8, 3 say  'B or <backspace>'
  @  9, 3 say  'M memory'
  @ 10, 3 say  'R recall'
  @ 11, 3 say  'C clear'
  @ 12, 3 say  'S change sign'
  @ 13, 3 say  'E return result'
  @ 14, 3 say  'Q or <esc> quit'
  @ 16, 1 say  'Any key to continue'
  wait
  popwind(wnb)
  set say video to vrev
return
*
* draw calculator with large buttons
*
procedure bgcalc
  frame(1)
  window(rw, cl, (rw + 17), (cl + 22))
  erase

  @ 1,  0,  1, 22     | draw a line under display

  @  2,  0,  4,  4     |1st row of buttons from top
  @  2,  6,  4, 10
  @  2, 12,  4, 16
  @  2, 18,  4, 22

  @  5,  0,  7,  4     |2nd row of buttons from top
  @  5,  6,  7, 10
  @  5, 12,  7, 16
  @  5, 18,  7, 22

  @  8,  0, 10,  4    |3rd row of buttons
  @  8,  6, 10, 10
  @  8, 12, 10, 16
  @  8, 18, 10, 22

  @ 11,  0, 13,  4    |4th row of buttons
  @ 11,  6, 13, 10
  @ 11, 12, 13, 16
  @ 11, 18, 13, 22

  @ 14,  0, 16,  4    |5th row of buttons
  @ 14,  6, 16, 10
  @ 14, 12, 16, 16
  @ 14, 18, 16, 22

  set say video to vrev

  @  3,  2 say '7'    | put # and symbols on the buttons
  @  3,  8 say '8'
  @  3, 14 say '9'
  @  3, 20 say '+'

  @  6,  2 say '4'
  @  6,  8 say '5'
  @  6, 14 say '6'
  @  6, 20 say '-'

  @  9,  2 say '1'
  @  9,  8 say '2'
  @  9, 14 say '3'
  @  9, 20 say '*'

  @ 12,  2 say '0'
  @ 12,  8 say '.'
  @ 12, 14 say 'S'
  @ 12, 20 say '/'

  @ 15,  2 say 'M'
  @ 15,  8 say 'R'
  @ 15, 14 say 'C'
  @ 15, 20 say '%'

  @ 17,  4 say 'F1 / H for help'
return
*
* draw calculator with small buttons
*
procedure smcalc
  frame(1)
  window(rw, cl, (rw + 12), (cl + 16))
  erase

  @ 1,  0,  1, 16     | line under display

  set say video to vrev

  @  2,  1 say ' 7 '  | display # and symbols
  @  2,  5 say ' 8 '
  @  2,  9 say ' 9 '
  @  2, 13 say ' + '

  @  4,  1 say ' 4 '
  @  4,  5 say ' 5 '
  @  4,  9 say ' 6 '
  @  4, 13 say ' - '

  @  6,  1 say ' 1 '
  @  6,  5 say ' 2 '
  @  6,  9 say ' 3 '
  @  6, 13 say ' * '

  @  8,  1 say ' 0 '
  @  8,  5 say ' . '
  @  8,  9 say ' S '
  @  8, 13 say ' / '

  @ 10,  1 say ' M '
  @ 10,  5 say ' R '
  @ 10,  9 say ' C '
  @ 10, 13 say ' % '

  @ 12,  1 say 'F1 / H for Help'
return
